@@ -1,11 +1,13 @@
 from typing import List, Optional
 
+import time
 import click
 import base64
 import fnmatch
 import json
 import boto3
 import subprocess
+import sys
 from dataclasses import dataclass
 
 from mypy_boto3_ecr import ECRClient
@@ -17,6 +19,8 @@ from mypy_boto3_ecr.type_defs import RepositoryTypeDef
 class Context:
     client: ECRClient
     registry_id: str
+    override_os: str
+    override_arch: str
 
 
 @dataclass()
@@ -24,13 +28,24 @@ class MirroredRepo:
     upstream_image: str
     repository_uri: str
     upstream_tags: List[str]
+    ignore_tags: List[str]
 
 
 @click.group()
-@click.option("--registry-id", help="The registry ID. This is usually your AWS account ID.")
+@click.option(
+    "--registry-id", help="The registry ID. This is usually your AWS account ID."
+)
 @click.option("--role-arn", help="Assume a specific role to push to AWS")
+@click.option(
+    "--override-os", default="linux", help="Specify the OS of images, default to \"linux\""
+)
+@click.option(
+    "--override-arch",
+    default="amd64",
+    help="Specify the ARCH of images, default to \"amd64\". If set to \"all\" - all architectures will be synced"
+)
 @click.pass_context
-def cli(ctx, registry_id, role_arn):
+def cli(ctx, registry_id, role_arn, override_os, override_arch):
     client = boto3.client("ecr")
     # Assume a role, if required:
     if role_arn:
@@ -49,7 +64,12 @@ def cli(ctx, registry_id, role_arn):
             aws_secret_access_key=tmp_secret_key,
             aws_session_token=security_token,
         )
-    ctx.obj = Context(client=client, registry_id=registry_id)
+    ctx.obj = Context(
+        client=client,
+        registry_id=registry_id,
+        override_os=override_os,
+        override_arch=override_arch,
+    )
 
 
 @cli.command()
@@ -76,6 +96,7 @@ def copy(ctx, source, destination_repository):
             upstream_image=upstream_image,
             upstream_tags=[upstream_tag],
             repository_uri=destination_repository,
+            ignore_tags=[]
         )
     ]
     copy_repositories(ctx.obj.client, ctx.obj.registry_id, repositories)
@@ -105,8 +126,9 @@ def ecr_login(client: ECRClient, registry_id: str) -> str:
     ).decode()
 
 
+@click.pass_context
 def copy_repositories(
-    client: ECRClient, registry_id: str, repositories: List[MirroredRepo]
+    ctx, client: ECRClient, registry_id: str, repositories: List[MirroredRepo]
 ):
     """
     Perform the actual, concurrent copy of the images
@@ -116,23 +138,25 @@ def copy_repositories(
     items = [
         (repo, tag)
         for repo in repositories
-        for tag in find_tags_to_copy(repo.upstream_image, repo.upstream_tags)
+        for tag in find_tags_to_copy(repo.upstream_image, repo.upstream_tags, repo.ignore_tags)
     ]
     click.echo(f"Beginning the copy of {len(items)} images")
 
-    with ThreadPoolExecutor() as pool:
+    with ThreadPoolExecutor(max_workers=4) as pool:
         # This code aint' beautiful, but whatever 🤷‍
         pool.map(
             lambda item: copy_image(
+                ctx.obj,
                 f"{item[0].upstream_image}:{item[1]}",
                 f"{item[0].repository_uri}:{item[1]}",
                 token,
+                sleep_time=1,
             ),
             items,
         )
 
 
-def copy_image(source_image, dest_image, token):
+def copy_image(ctx: Context, source_image, dest_image, token, sleep_time):
     """
     Copy a single image using Skopeo
     """
@@ -144,32 +168,54 @@ def copy_image(source_image, dest_image, token):
         "copy",
         f"docker://{source_image}",
         f"docker://{dest_image}",
-        "--override-os=linux",
+        f"--override-os={ctx.override_os}",
     ]
+    if ctx.override_arch == "all":
+        args = args + [f"--multi-arch=all"]
+    else:
+        args = args + [f"--override-arch={ctx.override_arch}"]
     args_with_creds = args + [f"--dest-creds={token}"]
     try:
-        subprocess.check_output(args_with_creds)
+        subprocess.check_output(args_with_creds, stderr=subprocess.STDOUT)
     except subprocess.CalledProcessError as e:
         click.secho(f'{" ".join(args)} raised an error: {e.returncode}', fg="red")
-        click.secho(f'Last output: {e.output[100:]}', fg="red")
+        click.secho(f"Last output: {e.output[100:]}", fg="red")
+
+    time.sleep(sleep_time)
 
 
-def find_tags_to_copy(image_name, tag_patterns):
+@click.pass_context
+def find_tags_to_copy(ctx, image_name, tag_patterns, ignore_tags):
     """
     Use Skopeo to list all available tags for an image
     """
-    output = subprocess.check_output(
-        ["skopeo", "list-tags", f"docker://{image_name}", "--override-os=linux"]
-    )
+    cmd = ["skopeo",
+           "list-tags",
+           f"docker://{image_name}",
+           f"--override-os={ctx.obj.override_os}"]
+    if ctx.obj.override_arch != "all":
+        cmd = cmd + [f"--override-arch={ctx.obj.override_arch}"]
+    output = subprocess.check_output(cmd)
     all_tags = json.loads(output)["Tags"]
 
-    if not tag_patterns:
-        return all_tags
+    def does_match(tag):
+        any_matches = any(fnmatch.fnmatch(tag, pattern) for pattern in tag_patterns)
+        any_negates = any(fnmatch.fnmatch(tag, pattern) for pattern in ignore_tags)
+
+        # Copy all tags if only ignored tags are given
+        if not tag_patterns and not any_negates:
+            return True
+
+        # Else if something has negated it, skip
+        if any_negates:
+            return False
+
+        return any_matches
 
     yield from (
         tag
         for tag in all_tags
-        if any(fnmatch.fnmatch(tag, pattern) for pattern in tag_patterns)
+        if does_match(tag)
     )
 
 
@@ -191,7 +237,12 @@ def find_repositories(client: ECRClient, registry_id: str):
         if "upstream-image" in tags_dict:
             return MirroredRepo(
                 upstream_image=tags_dict["upstream-image"],
-                upstream_tags=tags_dict.get("upstream-tags", "").replace("+", "*").split("/"),
+                upstream_tags=tags_dict.get("upstream-tags", "")
+                .replace("+", "*")
+                .split("/"),
+                ignore_tags=tags_dict.get("ignore-tags", "")
+                .replace("+", "*")
+                .split("/"),
                 repository_uri=repo["repositoryUri"],
             )
 
